/* eslint-disable max-len */

/**
 * Type of **GrapeCity.ActiveReports.Designer** object exported by **web-designer.js** module.
 */

type GlobalDesignerAPI = {
	/**
	 * Renders **Web Designer** to **<div>** element with id **designerElementId** using the specified **DesignerOptions** object.
	 *
	 * **Example:**
	 * ```javascript
	 * GrapeCity.ActiveReports.WebDesigner.create('designer-id', designerSettings);
	 * ```
	 *
	 * @param selector string
	 * @param settings DesignerSettings object
	 * @returns DesignerAPI of this instance
	 */
	create: (selector: string, settings: DesignerSettings) => Promise<DesignerAPI>;

	/**
	 * Returns DesignerAPI of previously created instance of **WebDesigner**
	 */
	apiOf: (instanceId: string) => DesignerAPI | undefined;

	/**
	 * Returns DesignerAPI of previously created instance of **WebDesigner**
	 */
	addLanguage: (lng: string, resources: ResourceBundle[]) => void;
};

type ResourceBundle = {
	/** Language */
	lng: string;

	/** Namespace */
	ns: string;

	/** Localization resources */
	resources: Record<string, any>;
};

/**
 * Type of object returned by functions from **GlobalDesignerAPI**
 */
type DesignerAPI = {

	/** Application-related APIs */
	app: {

		/** Contains documentation links and application info */
		about: {

			/** Designer-related documentation links */
			help: {
				/** General documentation */
				general: { text: string; url: string };
				/** Report items transformation info */
				transformation: { text?: string; url: string };
			};

			/** Application title */
			applicationTitle: string;
			/** A compact version of the application title */
			applicationTitleCompact: string;

			/** Application version */
			applicationVersion: string;
			/** Designer Core version an application is based on */
			coreVersion: string;
		}

		/**
		 * Returns focus to Designer. Focus may be lost when plug-in components are opened/closed.\
		 * Returning focus is essential to continue using Designer hotkeys like Ctrl+Z (undo), Ctrl+Y (redo), etc.
		 *
		 * **Example:**
		 * ```javascript
		 * GrapeCity.ActiveReports.Designer.apiOf('designer-id').app.focus();
		 * ```
		 */
		focus: () => void;
	};

	/**
	 * This object includes functions allowing to create/open/save report, etc.
	 */
	documents: DocumentsAPI;


	/**
	 * Allows to utilize built-in notifications system
	 */
	notifications: NotificationsAPI;
};

type DocumentsAPI = {
	/**
	 * Indicates whether report has unsaved changes.
	 *
	 * **Example:**
	 * ```javascript
	 * var hasUnsavedChanges = GrapeCity.ActiveReports.Designer.apiOf('designer-id').documents.hasUnsavedChanges();
	 * if (hasUnsavedChanges) console.log('Currently edited report has unsaved changes.');
	 * ```
	 */
	hasUnsavedChanges: () => boolean;

	/**
	 * Indicates whether report was saved before at least once
	 */
	isNew: () => boolean;

	/**
	 * Returns information about the currently edited report.
	 *
	 * **Example:**
	 * ```javascript
	 * var reportInfo = GrapeCity.ActiveReports.Designer.apiOf('designer-id').documents.info();
	 * console.log(`Report "${reportInfo.name}" is currently edited.`);
	 * ```
	 */
	info: () => CurrentDocumentInfo;

	/**
	 * Creates a new report to be edited in Designer using the specified **CreateReportOptions** object.
	 *
	 * **Example:**
	 * ```javascript
	 * var api = GrapeCity.ActiveReports.Designer.apiOf('designer-id');
	 * api.documents.create().then(() => {
	 * 		console.log('An empty RDL report is created.');
	 * })
	 * ```
	 *
	 * @param options CreateReportOptions object
	 */
	create: (options?: CreateDocumentOptions) => Promise<CreateDocumentInfo>;

	/**
	 * Shows open report dialog
	 */
	open: () => void;

	/**
	 * Opens an existing report to be edited in Designer with specified id.
	 * Optionally you can pass name and content, else it will be loaded from server.
	 *
	 * **Example:**
	 * ```javascript
	 * var api = GrapeCity.ActiveReports.Designer.apiOf('designer-id');
	 * api.documents.openById('MyReport.rdlx', { platform: 'rdlx', type: 'report', subType: 'cpl'}).then(() => {
	 *   console.log('An existing report "MyReport.rdlx" is opened.');
	 * });
	 * var reportContent = { Type: 'report', Body: { ReportItems: [ {Type: 'textbox', Name: 'TextBox1', Width: '5in', Height: '1in' } ] }};
	 * api.documents.openById('NewReport.rdlx', { platform: 'rdlx', type: 'report', subType: 'cpl'}, 'NewReport', reportContent).then(() => {
	 *   console.log('New report with one textbox created and opened.');
	 * });
	 * ```
	 *
	 * @param options OpenReportOptions object
	 */
	openById: (id: string, type: SupportedDocumentType, name?: string, content?: any) => Promise<OpenDocumentInfo>;

	/**
	 *  Saves the report currently edited in Designer, if report is new, then "Save As" dialog will be opened.
	 */
	save: () => void;

	/**
	 * Opens "Save As" dialog
	 */
	saveAs: () => void;

	/**
	 * Saves the report currently edited in Designer using the specified **id**.
	 *
	 * **Example:**
	 * ```javascript
	 * var api = GrapeCity.ActiveReports.Designer.apiOf('designer-id');
	 * api.documents.saveById('MyReport.rdlx')
	 * ```
	 * @param options SaveReportOptions object
	 */
	saveById: (id?: string, name?: string) => Promise<SaveDocumentInfo>;

	/**
	 * Saves the report currently edited in Designer using the specified **name**.
	 *
	 * **Example:**
	 * ```javascript
	 * var api = GrapeCity.ActiveReports.Designer.apiOf('designer-id');
	 * api.documents.saveById('MyReport.rdlx')
	 * ```
	 * @param options SaveReportOptions object
	 */
	saveByName: (name: string) => Promise<SaveDocumentInfo>;

};

type RpxReportDocumentType = {
	platform: 'rpx';
	type: 'report';
};

type RdlxFplReportDocumentType = {
	platform: 'rdlx';
	type: 'report';
	subType: 'fpl';
};

type RdlxCplReportDocumentType = {
	platform: 'rdlx';
	type: 'report'
	subType: 'cpl';
};

type RdlxReportDocumentType = RdlxCplReportDocumentType | RdlxFplReportDocumentType;

/**
 * Type of documents supported by the AR Web Designer
 */
type SupportedDocumentType = RpxReportDocumentType | RdlxReportDocumentType;

type NotificationsAPI = {
	/**
	 * Sends a notification of specified level, caption and content
	 * @param level Notification level. Determines color and icons used for the notifications
	 * @param caption Notification caption. Shown by default when notification pops up, then used as a title in Notification Details view
	 * @param content Notification content. Only visible when Notification Details are open
	 */
	send: (level: 'info' | 'warning' | 'error', caption: string, content?: string) => void;

	/**
	 * Sends a general notification. Can be used to notify when any user-initiated action is complete
	 * @param caption Notification caption
	 * @param text Optional notification content
	 */
	info: (caption: string, text?: string) => void;

	/**
	 * Sends an error notification
	 * @param caption Error caption
	 * @param errorText Optional error details
	 */
	error: (caption: string, errorText?: string) => void;

	/**
	 * Sends a warning notification
	 * @param caption Warning caption
	 * @param warningText Optional warning details
	 */
	warning: (caption: string, warningText?: string) => void;

	/**
	 * Dismisses all the notifications
	 */
	dismissAll: () => void;
};

type Font = { label: string; value: string };
type FontHeader = { header: string };

type DesignerSettings = {
	/**
	 * Unique identifier for the Designer instance. Required if there are multiple instances on the same page
	 * @default 'ar-web-designer'
	 */
	instanceId: string;

	/**
	 * Specifies language to use for the Designer.\
	 * If **language** is not specified, browser preferences are used
	 *
	 * **Example:**
	 * ```javascript
	 * designerSettings.language = 'ja';
	 * ```
	 * @default 'en'
	 */
	language: string;

	/**
	 * Specifies the list of fonts displayed in **Font** property editors all over Designer.\
	 * Supports plain strings, label-value pairs, headers and splitters
	 *
	 * If not specified, the default list of fonts is used:
	 *	'Arial', 'Arial Black', 'Comic Sans MS', 'Courier New', 'Geneva', 'Georgia', 'Helvetica',
	 *	'Impact', 'Lucida Console', 'Meiryo', 'Meiryo UI', 'MingLiU', 'MingLiU-ExtB', 'MS Gothic', 'MS Mincho',
	 *	'MS PGothic', 'MS PMincho', 'MS Song', 'MS UI Gothic', 'NSimSun', 'Osaka', 'PMingLiU', 'PMingLiU-ExtB',
	 *	'SimSun', 'SimSun-ExtB', 'Song', 'Tahoma', 'Times New Roman', 'Trebuchet MS', 'Verdana', 'Yu Gothic'.
	 *
	 * **Example:**
	 * ```javascript
	 * designerSettings.fonts = [{ header: 'Questionable Choice' }, { label: 'Pretty Font', value: 'Comic Sans MS' }, { header: '' }, 'Arial', 'Courier New', 'Times New Roman'];
	 * ```
	 */
	fonts: (string | Font | FontHeader)[];

	/**
	 * Specifies the list of supported image mime-types
	 */
	imageMimeTypes: string[];

	/**
	 * Specifies the default measurement units
	 *
	 * **Example:**
	 * ```javascript
	 * designerSettings.units = 'cm';
	 * ```
	 * @default 'in'
	 */
	units: 'in' | 'cm';

	/**
	 * When **lockLayout** is enabled, it is only possible to modify properties of existing report items.\
	 * I.e., adding a new report item or deleting of an existing one is not possible as well as other operations that modify report layout structure.\
	 *
	 * **Example:**
	 * ```javascript
	 * designerSettings.lockLayout = true;
	 * ```
	 * @default false
	 */
	lockLayout: boolean;

	/**
	 * Document to open on app startup, if not set new empty document will be created
	 */
	document?: {
		id: string;
		type: SupportedDocumentType;
	};

	/**
	 * When **storeUnsavedReport** is **enabled**, the last unsaved report can be restored if browser tab or browser itself gets accidentally closed.\
	 * In case **storeUnsavedReport** is **disabled**, the aforementioned functionality is not available
	 *
	 * **Example:**
	 * ```javascript
	 * designerSettings.storeUnsavedReport = false;
	 * ```
	 * @default true
	 */
	storeUnsavedReport: boolean;

	/**
	 * When **storeUserPreferences** is **enabled**, user preferences will be saved to a browser storage.\
	 * In case **storeUnsavedReport** is **disabled**, the aforementioned functionality is not available
	 *
	 * **Example:**
	 * ```javascript
	 * designerSettings.storeUserPreferences = false;
	 * ```
	 * @default true
	 */
	storeUserPreferences: boolean;

	/**
	 * By default, focused elements (like buttons) are highlighted only for a small period of time after **Tab** key was pressed.\
	 * This settings makes focused elements permanently highlighted.
	 * @default false
	 */
	disableFocusTimer: boolean;

	/**
	 * Disable usage of system clipboard. Copy-paste between designer instances will work only in the same browser in the same domain
	 * @default false
	 */
	disableSystemClipboard: boolean;

	/** Editor settings */
	editor: {

		/** Rulers settings */
		rulers?: {

			/** Specifies whether **Rulers** needs to be shown by default */
			visible: boolean;

			/** Specifies a snapStep value. By default snapStep = { in: 0.25, cm: 0.5 } */
			snapStep?: {
				in: number;
				cm: number;
			};
		};

		/** Specifies default grid size. If units = 'cm', value = 0.5cm by default. Else, value = 0,25in by default */
		gridSize?: string;

		/** Specifies if grid must be shown or hidden by default */
		showGrid?: boolean;

		/**
		 * Specifies default value for the snapToGrid option
		 * @default false
		 */
		snapToGrid?: boolean;

		/**
		 * Specifies default value for the snapToGuides option
		 * @default false
		 */
		snapToGuides?: boolean;
	};

	/** **App Bar** settings */
	appBar: {
		/**
		 * Specifies whether **App Bar** needs to be shown\
		 * @default true
		 */
		visible: boolean;


		/** **Save** button settings */
		saveButton: {
			/**
			 * Specifies whether **Save** button needs to be shown

			* **Example:**
			* ```javascript
			* designerSettings.appBar.saveButton.visible = false;
			* ```
			* @default true
			*/
			visible: boolean;
		};

		/** **Save As** button settings */
		saveAsButton: {
			/**
			 * Specifies whether **Save As** button needs to be shown

			* **Example:**
			* ```javascript
			* designerSettings.appBar.saveAsButton.visible = false;
			* ```
			* @default true
			*/
			visible: boolean;
		};

		/** **Open** button settings */
		openButton: {
			/**
			 * Specifies whether **Open** button needs to be shown

			* **Example:**
			* ```javascript
			* designerSettings.appBar.openButton.visible = true;
			* ```
			* @default false
			*/
			visible: boolean;
		};

		/** **Insert** tab settings */
		insertTab: {
			/**
			 * Specifies whether **Insert** tab needs to be shown in Designer application bar.\
			 * **Tool Box** and **Insert** tab are interchangeable
			 *
			 * **Example:**
			 * ```javascript
			 * designerSettings.appBar.insertTab.visible = true;
			 * ```
			 * @default false
			 */
			visible: boolean;
		};
	};

	/** **Menu** settings */
	menu: {
		/**
		 * Specifies whether **Main Menu** needs to be shown\
		 * @default true
		 */
		visible: boolean;

		/** **Tool Box** settings */
		toolBox: {
			/**
			 * Specifies whether left-side menu **Tool Box** needs to be shown
			 *
			 * **Example:**
			 * ```javascript
			 * designerSettings.menu.toolBox.visible = false;
			 * ```
			 * @default true
			 */
			visible: boolean;
		};

		/** **Document Explorer** settings */
		documentExplorer: {
			/**
			 * Specifies whether **Document Explorer** button needs to be shown
			 *
			 * **Example:**
			 * ```javascript
			 * designerSettings.menu.documentExplorer.visible = false;
			 * ```
			 * @default true
			 */
			visible: boolean;
		};

		/** **Group Editor** settings */
		groupEditor: {
			/**
			 * Specifies whether **Group Editor** button needs to be shown.\
			 *
			 * **Example:**
			 * ```javascript
			 * designerSettings.menu.groupEditor.visible = false;
			 * ```
			 * @default true
			 */
			visible: boolean;
		};
	};

	/** Status Bar settings */
	statusBar: {
		toggleUnitsButton: {
			/**
			 * Specifies whether **Units Button** button needs to be shown
			 * @default true
			 */
			visible: boolean;
		};
		toggleGridButton: {
			/**
			 * Specifies whether **Show Grid** toggle needs to be shown
			 * @default true
			 */
			visible: boolean;
		};
		gridSizeEditor: {
			/**
			 * Specifies whether **Grid Size** editor needs to be shown
			 * @default true
			 */
			visible: boolean;
		};
		rulersButton: {
			/**
			 * Specifies whether **Show Rulers** toggle needs to be shown
			 * @default true
			 */
			visible: boolean;
		};
		propertiesModeButton: {
			/**
			 * Specifies whether **Properties Mode** dropdown needs to be shown
			 * @default true
			 */
			visible: boolean;
		};
	};

	/** Property Grid settings */
	propertyGrid: {
		propertiesTab: {
			/**
			 * Specifies whether **Properties** tab needs to be shown
			 * @default true
			 */
			visible?: boolean;
		}

		/**
		 * Specifies default properties mode
		 */
		mode?: 'Basic' | 'Advanced';

		/**
		 * Specifies default properties sort mode
		 */
		sort?: 'categorized' | 'alphabetical';
	};

	/** Documents API settings */
	documents: {
		/** **File View** settings */
		fileView: {
			/**
			 * Specifies whether **File View** tab needs to be shown
			 *
			 * **Example:**
			 * ```javascript
			 * designerSettings.fileView.visible = false;
			 * ```
			 * @default true
			 */
			visible: boolean;
		};

		handlers: {
			/** An async handler, cancels saving process if returned false  */
			onBeforeSave?: (info: SaveDocumentInfo) => Promise<boolean>,
			/** A handler that is called when save is complete  */
			onAfterSave?: (info: SaveDocumentInfo) => void,

			/** An async handler, cancels opening process if returned false */
			onBeforeOpen?: () => Promise<boolean>,
			/** A handler that is called when open is complete */
			onAfterOpen?: () => void,

			/** An async handler, cancels create process if returned false */
			onBeforeCreate?: () => Promise<boolean>,
			/** A handler that is called when create is complete */
			onAfterCreate?: () => void,

			/** A handler that is triggered when report name/id is updated */
			onInfoUpdate?: (options: { name: string, id?: string }) => void;
		};
	};

	/** **Data**-related settings */
	data: {
		dataTab: {
			/**
			 * Specifies whether **Data** tab needs to be shown
			 * @default true
			 */
			visible: boolean;
		};

		/** **Data Sources** section settings */
		dataSources: {
			/**
			 * Specifies whether **Data Sources** section needs to be shown
			 *
			 * **Example:**
			 * ```javascript
			 * designerSettings.data.dataSources.visible = false;
			 * ```
			 * @default true
			 */
			visible: boolean;

			/**
			 * Specifies whether it is possible to modify (including add/edit/remove) data sources
			 *
			 * **Example:**
			 * ```javascript
			 * designerSettings.data.dataSources.canModify = true;
			 * ```
			 * @default false
			 */
			canModify: boolean;

			options: {
				/** Specifies the list of predefined data providers available in data source editor.\
				 * By default all the predefined providers are present.
				 *
				 * **Example:**
				 * ```javascript
				 * designerSettings.data.dataSources.options.predefinedProviders = ['SQL', 'JSON'];
				 * ```
				 */
				predefinedProviders: ('SQL' | 'OLEDB' | 'ODBC' | 'JSON' | 'CSV' | 'XML')[];

				/** Specifies the list of OLE DB providers available in data source editor.\
				 * By default 'Microsoft.Jet.OLEDB.4.0', 'SQLOLEDB.1', 'MSDataShape.1', 'MSDASQL.1' are present.
				 *
				 * **Example:**
				 * ```javascript
				 * designerSettings.data.dataSources.options.oleDbProviders = ['Microsoft.Jet.OLEDB.4.0', 'SQLOLEDB.1'];
				 * ```
				 */
				oleDbProviders: string[];

				/** Specifies the list of custom data providers available in data source editor.\
				 * By default there are no custom data providers present.
				 *
				 * **Example:**
				 * ```javascript
				 * designerSettings.data.dataSources.options.customProviders = [{ key: 'CDP', name: 'Custom Data Provider' }];
				 * ```
				 */
				customProviders: {
					/**
					* key - data provider identifier\
					* This value is used for **DataSource.ConnectionProperties.DataProvider** property.
					*/
					key: string;

					/**
					* name - data provider label\
					* This value is used as a friendly data provider label in UI.
					*/
					name: string;
				}[];
			};
		};

		dataSets: {
			/**
			 * Specifies whether **Data Sets** section needs to be shown.\
			 *
			 * **Example:**
			 * ```javascript
			 * designerSettings.data.dataSets.visible = false;
			 * ```
			 * @default true
			 */
			visible: boolean;

			/**
			 * Specifies whether it is possible to modify (including add/edit/remove) data sets.\
			 *
			 * **Example:**
			 * ```javascript
			 * designerSettings.data.dataSets.canModify = true;
			 * ```
			 * @default false
			 */
			canModify: boolean;
		};

		parameters: {
			/**
			 * Specifies whether **Parameters** section needs to be shown.\
			 *
			 * **Example:**
			 * ```javascript
			 * designerSettings.data.parameters.visible = false;
			 * ```
			 * @default true
			 */
			visible: boolean;

			/**
			 * Specifies whether it is possible to modify (including add/edit/remove) report parameters.\
			 *
			 * **Example:**
			 * ```javascript
			 * designerSettings.data.parameters.canModify = false;
			 * ```
			 * @default true
			 */
			canModify: boolean;
		};

		commonValues: {
			/**
			 * Specifies whether **Common Values** section needs to be shown.\
			 *
			 * **Example:**
			 * ```javascript
			 * designerSettings.data.commonValues.visible = false;
			 * ```
			 * @default true
			 */
			visible: boolean;
		};
	};

	/** RPX **Styles**-related settings */
	styles: {
		stylesTab: {
			/**
			 * Specifies whether **Styles** tab needs to be shown for RPX reports
			 * @default true
			 */
			visible: boolean;
		};

		stylesheet: {
			/**
			 * Specifies whether RPX report Stylesheet can be modified
			 * @default true
			 */
			canModify: boolean;
		};
	};

	/** Backend-related settings */
	server: {
		/**
		 * Specifies the base URL for Designer Server API.
		 * @default api
		 *
		 * **Example:**
		 * ```javascript
		 * designerSettings.server.url = 'api/designer';
		 * ```
		 */
		url: string;

		/**
		 * A special handler to modify requests
		 */
		onBeforeRequest?: (init: RequestInit) => RequestInit;
	};

	/** Document **Title** settings */
	title: {
		/**
		 * You can implement custom logic for updating browser tab's title when edited document gets updated in Designer
		 *
		 * **Example:**
		 * ```javascript
		 * designerSettings.title.onUpdate = ({ name, hasUnsavedChanges }) => `${name}${hasUnsavedChanges ? ' - Has Unsaved Changes!' : ''}`;
		 * ```
		 * @param options TitleInfo object
		 * @default undefined
		 */
		onUpdate?: (info: TitleInfo) => string;

		/**
		 * Specifies whether browser tab title can be updated
		 *
		 * **Example:**
		 * ```javascript
		 * designerSettings.title.disabled = false;
		 * ```
		 * @default false
		 */
		disabled: boolean;
	};

	/** Document **Preview** settings */
	preview: {
		/**
		 * Specifies whether **Preview** button needs to be shown
		 * @default true
		 */
		canPreview?: boolean;

		/**
		 * You can plug-in **Viewer** component by providing **openViewer** function\
		 *
		 * **Example:**
		 * ```javascript
		 * designerSettings.preview.openViewer = function(options) {
		 * 	// ... create viewer and open report
		 * };
		 * ```
		 *
		 * @param settings ViewerSettings object
		 */
		openViewer: (settings: ViewerSettings) => void;
	}

	/** RPX platform-specific settings. Must exist for the RPX reports to work */
	rpx: {
		initTemplates?: {
			imperialTemplates?: string[];
			metricTemplates?: string[];
		};
	};

	rdlx: {
		/**
		 * Specifies the expression syntax used in Designer:
		 * 'i11n' - interpolation syntax, 'rdl' is an "old" rdl expression syntax
		 *
		 * **Example:**
		 * ```javascript
		 * designerSettings.rdlx.expressionSyntax = 'rdl';
		 * ```
		 * @default 'i11n'
		 */
		expressionSyntax: 'i11n' | 'rdl';

		/**
		 * Filter for Expression Editor nodes
		 *
		 *	 **Example:**
		* ```javascript
		* 	designerSettings.rdlx.expressionEditorNodesFilter = (key) => {
		* 		if (key.includes('CommonValues')) return false;
		* 		if (key.includes('aggregate.aggregateIfWithScope')) return false;
		* 		return true;
		* 	}
		*/
		expressionEditorNodesFilter?: (key: string) => boolean;

		/**
		 * Specifies report items available and their order\
		 *
		 * Default items for CPL report: [ 'textbox', 'checkbox', 'container', 'line', 'shape', 'tableofcontents',
		 *	'image', 'list', 'table', 'tablix', 'chart', 'bullet', 'barcode', 'formattedtext',
		*	'richtext', 'sparkline', 'subreport', 'bandedlist', 'inputfield' ]
		*
		* Default items for FPL report: [ 'textbox', 'checkbox', 'container', 'line', 'shape', 'tableofcontents',
		*	'image', 'list', 'table', 'tablix', 'chart', 'bullet', 'barcode', 'formattedtext',
		*	'richtext', 'sparkline', 'subreport', 'overflowplaceholder', 'bandedlist', 'inputfield' ]
		*
		* **Example:**
		* ```javascript
		* designerSettings.rdlx.toolBoxContent = {
		* 		cpl: [ 'checkbox', 'container', 'textbox' ],
		* 		fpl: [ 'image', 'list', 'table', 'tablix', 'chart', 'bullet', 'barcode', 'formattedtext' ],
		* };
		* ```
		*/
		toolBoxContent?: {
			cpl: RdlxToolBoxItem[];
			fpl: RdlxToolBoxItem[];
		};

		/** Reports as a rdlx-json strings, report items from these reports will be used as a templates for creating new items
		 *
		 * imperialTemplates should include reports with units = 'in'
		 * metricTemplates should include reports with units = 'cm'
		 *
		 * Should use fpl-reports to set custom templates for specific fpl-reportItems, for example 'OverflowPlaceHolder'
		 * It is preferable to use cpl-reports to set custom templates for all other report items, as well as for pageHeader and pageFooter.
		 *
		 * For Report you can set ConsumeContainerWhitespace and Page-properties: BottomMargin, LeftMargin, RightMargin, TopMargin,	PageHeight,	PageWidth, ColumnSpacing
		 * For the rest of the report items, all properties are set, including, for example, the number of columns in the table or barcode default symbology
		 *
		 * Note: If in a subsequent report in the array there is the same reportItem as in the previous one, only last template for this reportItem will be set
		 *
		 * Also you can set multiple init templates for some report items. For this, you should add report with more then one report items the same type. Templates with the same names will be replaced.
		 * Use 'TemplateName' custom property of each report item to set a localized name of the template
		 * Use 'AllowWizard' custom property of a chart report item to allow use the chart wizard when this report item is created
		 * For example: ... ,"CustomProperties":[{"Name":"TemplateName","Value":"Doughnut Chart"}, {"Name":"AllowWizard","Value":"true"},...],...
		 *
		 *	 **Example:**
		* ```javascript
		*	 designerSettings.rdlx.initTemplates = {
		*	 	imperialTemplates: ['{"Name":"Report","Width":"5in","Layers":[{"Name":"default"}],"CustomProperties":[{"Name":"DisplayType","Value":"Page"},{"Name":"SizeType","Value":"Default"},{"Name":"PaperOrientation","Value":"Portrait"}],"Page":{"PageWidth":"8.5in","PageHeight":"11in","RightMargin":"1in","LeftMargin":"1in","TopMargin":"1in","BottomMargin":"1in","Columns":1,"ColumnSpacing":"0in"},"Body":{"Height":"0.25in","ReportItems":[{"Type":"textbox","Name":"TextBox1","CustomProperties":[],"CanGrow":true,"KeepTogether":true,"Style":{"PaddingLeft":"2pt","PaddingRight":"2pt","PaddingTop":"2pt","PaddingBottom":"2pt"},"Width":"5in","Height":"0.25in"}]}}'],
		*	 	metricTemplates: ['{"Name":"Report","Width":"10cm","Layers":[{"Name":"default"}],"CustomProperties":[{"Name":"DisplayType","Value":"Page"},{"Name":"SizeType","Value":"Default"},{"Name":"PaperOrientation","Value":"Portrait"}],"Page":{"PageWidth":"8.5in","PageHeight":"11in","RightMargin":"1in","LeftMargin":"1in","TopMargin":"1in","BottomMargin":"1in","Columns":1,"ColumnSpacing":"0in"},"Body":{"Height":"0.75cm","ReportItems":[{"Type":"textbox","Name":"TextBox1","CustomProperties":[],"CanGrow":true,"KeepTogether":true,"Style":{"PaddingLeft":"2pt","PaddingRight":"2pt","PaddingTop":"2pt","PaddingBottom":"2pt"},"Left":"0cm","Top":"0cm","Width":"10cm","Height":"0.75cm"}]}}'],
		* 	};
		* 	 * ```
		*/
		initTemplates?: {
			imperialTemplates?: string[];
			metricTemplates?: string[];
		};

		/**
		 * Customizable report item features
		 */
		reportItems: RdlxReportItemsSettings;

		/** Specifies additional styles to add to report item styles.
		 *
		 *	 **Example:**
		* ```javascript
		* designerSettings.rdlx.reportStyles = [
		*				{
		*					Bullet: [{
		*							id: 'c8aa4403-83ef-402b-a7da-032063cf625a',
		*							name: 'additionalBulletStyle1',
		*							content: {
		*								ValueColor: 'red',
		*								LabelFontColor: '=Theme.Colors!Dark1',
		*								LabelFontFamily: '=Theme.Fonts!MinorFont.Family',
		*								LabelFontSize: '=Theme.Fonts!MinorFont.Size',
		*								LabelFontStyle: '=Theme.Fonts!MinorFont.Style',
		*								TicksLineColor: '=Theme.Colors(1,0)'
		*							}
		*						},
		*					],
		*					List: [
		*						{	id: '3d2c3781-4eea-4ac3-8d50-636edd9328d5',
		*							name: 'additionalListStyle1',
		*							content: {},
		*						},
		*						{	id: '5b7b4e73-22e5-42ed-99c4-62840bdde79d',
		*							name: 'additionalListStyle2',
		*							content: {
		*								BackgroundColor: '=Theme.Colors!Accent1',
		*								Border: {
		*									Color: '=Theme.Colors(4,4)',
		*									Width: '1pt',
		*									Style: 'Solid',
		*								}
		*							}
		*						}]
		*				},
		*				{	ChartPalette: [{
		*							id: 'c8aa4403-83ef-402b-a7da-0320444',
		*							name: 'additionalChartPalette',
		*							content: ['red', '=Theme.Colors!Accent2', '=Theme.Colors!Accent3', '=Theme.Colors!Accent4', '=Theme.Colors!Accent5', '=Theme.Colors!Accent6', '=Theme.Colors(5,4)', '=Theme.Colors(5,5)', '=Theme.Colors(5,6)', '=Theme.Colors(5,7)', '=Theme.Colors(5,8)', '=Theme.Colors(5,9)']
		*						}]
		*				},
		*			];
		* ```
		*/
		reportStyles: ReportStyles[];
	};
};

type TitleInfo = {
	/** Document name */
	name: string;
	/** Indicates whether document has unsaved changes or not. */
	hasUnsavedChanges: boolean;
};

type RdlxToolBoxItem = 'textbox' | 'checkbox' | 'container' | 'line' | 'shape' | 'tableofcontents'
	| 'image' | 'list' | 'table' | 'tablix' | 'chart' | 'bullet' | 'barcode' | 'formattedtext'
	| 'richtext' | 'sparkline' | 'subreport' | 'overflowplaceholder' | 'bandedlist' | 'inputfield';

type RdlxReportItemsSettings = {
	/** Barcode features */
	barcode: {
		/**
		 * Limits the list of barcode symbologies available for creation.\
		 * By default all barcode symbologies supported by ActiveReports are available.
		 *
		 * **Example:**
		 * ```javascript
		 * designerSettings.rdlx.reportItemsFeatures.barcode.symbologies = ['Code_128_A', 'Code_128_B', 'Code_128_C'];
		 * ```
		 */
		symbologies?: RdlxBarcodeSymbology[];
	};

	/** Table features */
	table: {
		/**
		 * Specifies whether **Table Header** needs to be auto-filled when a field is dropped to **Table Details**.\
		 * For example, if **ProductName** field is dropped to **Details**, **Product Name** value is set to **Header**
		 *
		 * **Example:**
		 * ```javascript
		 * designerSettings.rdlx.reportItemsFeatures.table.autoFillHeader = false;
		 * ```
		 * @default true
		 */
		autoFillHeader?: boolean;

		/**
		 * Specifies whether **Table Footer** needs to be auto-filled when a field is dropped to **Table Details**.\
		 * For example, if **ProductName** field is dropped to **Details**, **=Count([ProductName])** value is set to **Footer**
		 *
		 * **Example:**
		 * ```javascript
		 * designerSettings.rdlx.reportItemsFeatures.table.autoFillFooter = true;
		 * ```
		 * @default false
		 */
		autoFillFooter?: boolean;

		/**
		 * Specifies whether vertical merge of cells is enabled within **Table Header**, **Details** and **Footer**
		 *
		 * **Example:**
		 * ```javascript
		 * designerSettings.rdlx.reportItemsFeatures.table.canMergeCellsVertically = false;
		 * ```
		 * @default true
		 */
		canMergeCellsVertically?: boolean;
	};

	/** Tablix features */
	tablix: {
		/**
		 * Specifies whether Tablix Wizard should hide cross-aggregates functionality
		 *
		 * **Example:**
		 * ```javascript
		 * designerSettings.rdlx.reportItemsFeatures.tablix.crossAggregates = false;
		 * ```
		 * @default true
		 */
		crossAggregates?: boolean,

		/**
		 * Specifies whether **Tablix Corner Cell** needs to be auto-filled when a field is dropped to **Tablix Row Group Cell**.\
		 * For example, if **ProductName** field is dropped to **Row Group Cell**, **Product Name** value is set to **Corner Cell**
		 *
		 * **Example:**
		 * ```javascript
		 * designerSettings.rdlx.reportItemsFeatures.tablix.autoFillCorner = false;
		 * ```
		 * @default true
		 */
		autoFillCorner?: boolean;

		/**
		 * Specifies whether Tablix Wizard is available for creating/editing Tablix
		 *
		 * **Example:**
		 * ```javascript
		 * designerSettings.rdlx.reportItemsFeatures.tablix.canUseWizard = false;
		 * ```
		 * @default true
		 */
		canUseWizard?: boolean;
	};
};

type RdlxBarcodeSymbology = 'Ansi39' | 'Ansi39x' | 'BC412' | 'Codabar' | 'Code_11' | 'Code_128_A' | 'Code_128_B' | 'Code_128_C' | 'Code_128auto'
	| 'Code_2_of_5' | 'Code_93' | 'Code25intlv' | 'Code39' | 'Code39x' | 'Code49' | 'Code93x' | 'DataMatrix' | 'EAN_13' | 'EAN_8' | 'EAN128FNC1'
	| 'GS1QRCode' | 'HIBCCode128' | 'HIBCCode39' | 'IATA_2_of_5' | 'IntelligentMail' | 'IntelligentMailPackage' | 'ISBN' | 'ISMN' | 'ISSN'
	| 'ITF14' | 'JapanesePostal' | 'Matrix_2_of_5' | 'MaxiCode' | 'MicroPDF417' | 'MicroQRCode' | 'MSI' | 'Pdf417' | 'Pharmacode' | 'Plessey'
	| 'PostNet' | 'PZN' | 'QRCode' | 'RM4SCC' | 'RSS14' | 'RSS14Stacked' | 'RSS14StackedOmnidirectional' | 'RSS14Truncated' | 'RSSExpanded'
	| 'RSSExpandedStacked' | 'RSSLimited' | 'SSCC_18' | 'Telepen' | 'UCCEAN128' | 'UPC_A' | 'UPC_E0' | 'UPC_E1';

type ReportItemStyle<T> = {
	id: string;
	name: string;
	content: T;
};

type BorderStyle = {
	Color?: string;
	Style?: string;
	Width?: string;
};

type CellStyleContent = TextBoxStyleContent & {
	LeftBorder?: BorderStyle;
	TopBorder?: BorderStyle;
	BottomBorder?: BorderStyle;
	RightBorder?: BorderStyle;
	TextAlign?: string;
	VerticalAlign?: string;
};

type TextBoxStyleContent = {
	Color?: string;
	BackgroundColor?: string;
	FontFamily?: string;
	FontSize?: string;
	FontStyle?: string;
	FontWeight?: string;
};

type ContainerStyleContent = {
	BackgroundColor?: string;
	Border?: BorderStyle;
};

type BulletStyleContent = {
	ValueColor?: string;
	TargetLineColor?: string;
	LabelFontColor?: string;
	LabelFontFamily?: string;
	LabelFontSize?: string;
	LabelFontStyle?: string;
	TicksLineColor?: string;
};

type SparklineStyleContent = {
	LineColor?: string;
	FillColor?: string;
	GradientEndColor?: string;
	MarkerColor?: string;
	RangeFillColor?: string;
	RangeGradientEndColor?: string;
};

type TableStyleContent = {
	Header: { Rows: CellStyleContent[] };
	Details: { Rows: CellStyleContent[]; AlternatingExpression: string };
	Footer: { Rows: CellStyleContent[] };
	TableGroups: {
		Styles: {
			Header: { Rows: CellStyleContent[] };
			Footer: { Rows: CellStyleContent[] };
		}[];
	};
	Border?: BorderStyle;
};

type TableOfContentsStyleContent = {
	BackgroundColor?: string;
	Border?: BorderStyle;
	Levels: {
		Color?: string;
		PaddingLeft?: string;
		FontFamily?: string;
		FontSize?: string;
		FontStyle?: string;
		FontWeight?: string;
	};
};

type ChartPaletteContent = string[];

type ReportStyles = {
	Bullet: ReportItemStyle<BulletStyleContent>;
	CheckBox: ReportItemStyle<TextBoxStyleContent>;
	FormattedText: ReportItemStyle<ContainerStyleContent>;
	InputField: ReportItemStyle<TextBoxStyleContent>;
	List: ReportItemStyle<ContainerStyleContent>;
	Rectangle: ReportItemStyle<ContainerStyleContent>;
	RichText: ReportItemStyle<TextBoxStyleContent>;
	Shape: ReportItemStyle<ContainerStyleContent>;
	Sparkline: ReportItemStyle<SparklineStyleContent>;
	Table: ReportItemStyle<TableStyleContent>;
	TableOfContents: ReportItemStyle<TableOfContentsStyleContent>;
	Tablix: ReportItemStyle<TableStyleContent>;
	TextBox: ReportItemStyle<TextBoxStyleContent>;
	ChartPalette: ReportItemStyle<ChartPaletteContent>;
};

type ViewerSettings = {
	/** Host element's **id** to render Viewer to */
	element: string;

	/** Application title passed by Designer */
	applicationTitle: string;

	/** Information on the document to-be-previewed */
	documentInfo: {
		/** Document id */
		id: string;

		/** Document content in JSON format that can be useful for viewers with in-browser rendering. */
		content: unknown;

		/** Document name */
		name: string;

		/** Specifies whether the document to-be-previewed is an existing document saved on server side. */
		temporary: boolean;
	};
};

type CreateDocumentOptions = {
	/**
	 * Name of the document. If not specified default 'Untitled' name will be used
	 */
	name?: string;

	/**
	 * Type of the document to create. If not specified RDL report will be created
	 */
	type?: SupportedDocumentType;

	/**
	 * Template to use for the document
	 */
	template?: DocumentTemplate;

	/**
	 * A list of RDL Data Sets to use with the template
	 *
	 * **NOTE:** This won't work with RPX reports
	 */
	dataSetsInfo?: {
		/** data set id */
		id: string;

		/** data set name */
		name: string;
	}[];
};

type CreateDocumentInfo = {
	/**
	 * Document type
	 */
	type: SupportedDocumentType;

	/**
	 * Name of the document
	 */
	name: string;

	/**
	 * Document template
	 */
	template?: DocumentTemplate;

	/**
	 * undefined in onBefore**** handlers.
	 * Should be defined in onAfter****
	 */
	success?: boolean;
};

type CurrentDocumentInfo = {
	/**
	 * Document id
	 *
	 * If an existing report is edited, **id** is defined.\
	 * Otherwise, if a new report is edited, **id** is **null**.
	 */
	id: string | null;

	/** Document name */
	name: string;

	/**
	 * Document type
	 */
	type: SupportedDocumentType;
};

type DocumentTemplate = {
	/** Document template id */
	id?: string;

	/** Document template content in JSON format that can be useful in case you would like to create a non-empty new report. */
	content?: unknown;
};

type OpenDocumentOptions = {
	/** template info - if it is specified for report creation, either **templateInfo.id** or **templateinfo.content** needs to be defined */
	templateInfo?: DocumentTemplate;

	/**
	 * A list of RDL Data Sets to use with the template
	 *
	 * **NOTE:** This won't work with RPX reports
	 */
	dataSetsInfo?: {
		/** data set id */
		id: string;

		/** data set name */
		name: string;
	}[];

	/**
	 * Name for new report, if not specified default 'Untitled' name will be used
	 */
	name?: string;

	/**
	 * Type of the document to create. If not specified RDL report will be created
	 */
	type?: SupportedDocumentType;
};

type SaveDocumentInfo = {
	/**
	 * Type of the document
	 */
	type: SupportedDocumentType;

	/**
	 * If an existing document is to be overwritten on saving,
	 * the correct id should be specified explicitly
	 */
	id?: string;

	/**
	  * The correct name must be specified explicitly
	  */
	name: string;

	/**
	  * Indicates that a new document is being saved for the first time
	  */
	isFirstSave: boolean;

	/**
	  * undefined in onBefore**** handlers.
	  * Should be defined in onAfter****
	  */
	success?: boolean;
};

type OpenDocumentInfo = {
	/**
	 * Document type
	 */
	type: SupportedDocumentType;

	/**
	 * Document id
	 */
	id: string;

	/**
	 * Document name
	 */
	name: string;

	/**
	 * undefined in onBefore**** handlers.
	 * Should be defined in onAfter****
	 */
	success?: boolean;
};
